﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using Microsoft.ServiceBus;
using Microsoft.ServiceBus.Messaging;

namespace Nardax.Azure.ServiceBus
{
    public class SubscriptionAdapter : SubscriptionAdapter<BrokeredMessage>, ISubscriptionAdapter
    {
        public SubscriptionAdapter(SubscriptionAdapterDescription subscriptionAdapterDescription)
            : base(subscriptionAdapterDescription)
        { }
    }

    public class SubscriptionAdapter<T> : ISubscriptionAdapter<T>
    {
        private readonly ISubscriptionClientFactory _subscriptionClientFactory;
        private readonly BlockingCollection<BrokeredMessage> _brokeredMessages;
        private bool _isInitialized;
        private readonly object _initializationLock;
        private readonly ISubscriptionClient[] _subscriptionClients;

        public SubscriptionAdapter(SubscriptionAdapterDescription<T> subscriptionAdapterDescription)
            : this(subscriptionAdapterDescription, new SubscriptionClientFactory())
        {
        }

        public SubscriptionAdapter(SubscriptionAdapterDescription<T> subscriptionAdapterDescription, ISubscriptionClientFactory subscriptionClientFactory)
        {
            Guard.ArgumentNotNull(subscriptionAdapterDescription, "subscriptionAdapterDescription");

            SubscriptionAdapterDescription = subscriptionAdapterDescription;
            _subscriptionClientFactory = subscriptionClientFactory;
            _brokeredMessages = new BlockingCollection<BrokeredMessage>();
            _initializationLock = new object();
            _subscriptionClients = new ISubscriptionClient[SubscriptionAdapterDescription.ConnectionStrings.Length];
        }

        public void Initialize()
        {
            try
            {
                InitializeInternal();
            }
            catch (Exception ex)
            {
                var initializationException = new ServiceBusAdapterInitializationException("SubscriptionAdapter failed to initialize.", ex);
                throw initializationException;
            }
        }

        private void InitializeInternal()
        {
            if (_isInitialized)
            {
                return;
            }

            lock (_initializationLock)
            {
                if (_isInitialized)
                {
                    return;
                }

                for (var i = 0; i < SubscriptionAdapterDescription.ConnectionStrings.Length; i++)
                {
                    TryCreateSubscription(i);
                    _subscriptionClients[i] = CreateSubscriptionClient(i);
                }

                _isInitialized = true;
            }
        }

        private void TryCreateSubscription(int i)
        {
            var namespaceManager = NamespaceManager.CreateFromConnectionString(SubscriptionAdapterDescription.ConnectionStrings[i]);
            if (!namespaceManager.TopicExists(SubscriptionAdapterDescription.TopicDescription.Path))
            {
                try
                {
                    SubscriptionAdapterDescription.RetryPolicy.Execute(() => namespaceManager.CreateTopic(SubscriptionAdapterDescription.TopicDescription));
                }
                catch (MessagingEntityAlreadyExistsException)
                {
                    // I am handling this exception because it is telling me that the topic I am trying to create already
                    // exists. This is possible because other adapters may have created it just after I checked to see
                    // if it was already there.
                }
            }

            if (!namespaceManager.SubscriptionExists(SubscriptionAdapterDescription.TopicDescription.Path, SubscriptionAdapterDescription.SubscriptionDescription.Name))
            {
                try
                {
                    if (SubscriptionAdapterDescription.RuleDescription == null)
                    {
                        SubscriptionAdapterDescription.RetryPolicy.Execute(() => namespaceManager.CreateSubscription(SubscriptionAdapterDescription.SubscriptionDescription));
                    }
                    else
                    {
                        SubscriptionAdapterDescription.RetryPolicy.Execute(() => namespaceManager.CreateSubscription(SubscriptionAdapterDescription.SubscriptionDescription,
                                                                                                                      SubscriptionAdapterDescription.RuleDescription));
                    }
                }
                catch (MessagingEntityAlreadyExistsException)
                {
                    // I am handling this exception because it is telling me that the subscription I am trying to create already
                    // exists. This is possible because other adapters may have created it just after I checked to see
                    // if it was already there.
                }
            }
        }

        private ISubscriptionClient CreateSubscriptionClient(int i)
        {
            var subscriptionClient = _subscriptionClientFactory.CreateSubscriptionClient(SubscriptionAdapterDescription.ConnectionStrings[i],
                                                                                   SubscriptionAdapterDescription.TopicDescription.Path,
                                                                                   SubscriptionAdapterDescription.SubscriptionDescription.Name);
            
            subscriptionClient.PrefetchCount = SubscriptionAdapterDescription.PrefetchCount;

            return subscriptionClient;
        }

        public IEnumerable<T> GetConsumingEnumerable()
        {
            Initialize();

            // I am making the assumption that there is no data in the _brokeredMessages collection
            // and that it is necessary to seed the collection with at least one message in order
            // to get the _brokeredMessages enumeration to start its cycle. In the worst case there 
            // should only be one message per subscription client in the _brokeredMessages collection and 
            // it should be assumed that those messages have timed out.
            GetMessages();

            foreach (var brokeredMessage in _brokeredMessages.GetConsumingEnumerable())
            {
                if (!_brokeredMessages.IsAddingCompleted && _brokeredMessages.Count < 1)
                {
                    GetMessages();
                }

                if (brokeredMessage == null)
                {
                    // brokered message is only null when there are no messages in the subscription
                    // that need to be processed. This is not an exceptional case and I simply need
                    // to allow the loop to continue and eventually pull down new messages.
                    continue;
                }

                var brokeredMessageContainer = SubscriptionAdapterDescription.BrokeredMessageContainerFactory.CreateBrokeredMessageContainer(brokeredMessage);

                yield return brokeredMessageContainer;
            }
        }

        public void StopReceiving()
        {
            _brokeredMessages.CompleteAdding();
        }

        public void Close()
        {
            StopReceiving();

            foreach (var subscriptionClient in _subscriptionClients)
            {
                subscriptionClient.Close();
            }
        }

        /// <summary>
        /// all operation must result in a message being added to the _brokeredMessages collection
        /// even if that operation results in an exception... especially when it results in an exception
        /// if each call to GetMessages doesn't add something into the _BrokeredMessages collection
        /// the loop on the consuming enumerable will hang indefinitely
        /// </summary>
        private void GetMessages()
        {
            for (var i = 0; i < _subscriptionClients.Length; i++)
            {
                var sc = _subscriptionClients[i];
                try
                {
                    sc.BeginReceive(SubscriptionAdapterDescription.ServerWaitTime, ar =>
                        {
                            try
                            {
                                // if the call to EndReceive times out it will return null and not thrown an exception
                                var message = sc.EndReceive(ar);
                                AddMessageToCollection(message);
                            }
                            catch (Exception ex)
                            {
                                AddMessageToCollection(null);
                                OnReceiveFailed(ex);
                            }
                        }, null);
                }
                catch (Exception ex)
                {
                    if (ex is OperationCanceledException)
                    {
                        // Client entity has been closed or aborted so I am going to replace it
                        _subscriptionClients[i] = CreateSubscriptionClient(i);
                    }

                    AddMessageToCollection(null);
                    OnReceiveFailed(ex);
                }
            }
        }

        private void AddMessageToCollection(BrokeredMessage brokeredMessage)
        {
            if (!_brokeredMessages.IsAddingCompleted)
            {
                _brokeredMessages.Add(brokeredMessage);
            }
        }

        private void OnReceiveFailed(Exception exception)
        {
            if (ReceiveFailed != null)
            {
                ReceiveFailed(this, new ReceiveFailedEventArgs(exception));
            }
        }

        public event EventHandler<ReceiveFailedEventArgs> ReceiveFailed;

        public SubscriptionAdapterDescription<T> SubscriptionAdapterDescription { get; private set; }
    }
}