﻿using System;
using Microsoft.ServiceBus;
using Microsoft.ServiceBus.Messaging;

namespace Nardax.Azure.ServiceBus
{
    public class TopicAdapter : ITopicAdapter
    {
        private readonly ITopicClientFactory _topicClientFactory;
        private bool _isInitialized;
        private readonly object _initializationLock;
        private readonly ITopicClient[] _topicClients;
        private readonly object _startingIndexLock;
        private int _startingIndex;

        public TopicAdapter(TopicAdapterDescription topicAdapterDescription)
            : this(topicAdapterDescription, new TopicClientFactory())
        {
        }

        public TopicAdapter(TopicAdapterDescription topicAdapterDescription, ITopicClientFactory topicClientFactory)
        {
            Guard.ArgumentNotNull(topicAdapterDescription, "topicAdapterDescription");
            Guard.ArgumentNotNull(topicClientFactory, "topicClientFactory");

            TopicAdapterDescription = topicAdapterDescription;
            _topicClientFactory = topicClientFactory;
            _initializationLock = new object();
            _topicClients = new ITopicClient[TopicAdapterDescription.ConnectionStrings.Length];
            _startingIndexLock = new object();
        }

        public void Initialize()
        {
            try
            {
                InitializeInternal();
            }
            catch (Exception ex)
            {
                var initializationException = new ServiceBusAdapterInitializationException("TopicAdapter failed to initialize.", ex);
                throw initializationException;
            }
        }

        private void InitializeInternal()
        {
            if (_isInitialized)
            {
                return;
            }

            lock (_initializationLock)
            {
                if (_isInitialized)
                {
                    return;
                }

                for (var i = 0; i < TopicAdapterDescription.ConnectionStrings.Length; i++)
                {
                    TryCreateTopic(i);
                    _topicClients[i] = CreateTopicClient(i);
                }

                _isInitialized = true;
            }
        }

        private void TryCreateTopic(int index)
        {
            var namespaceManager = NamespaceManager.CreateFromConnectionString(TopicAdapterDescription.ConnectionStrings[index]);
            if (!namespaceManager.TopicExists(TopicAdapterDescription.TopicDescription.Path))
            {
                try
                {
                    TopicAdapterDescription.RetryPolicy.Execute(() => namespaceManager.CreateTopic(TopicAdapterDescription.TopicDescription));
                }
                catch (MessagingEntityAlreadyExistsException)
                {
                    // I am handling this exception because it is telling me that the topic I am trying to create already
                    // exists. This is possible because other adapters may have created it just after I checked to see
                    // if it was already there.
                }
            }
        }

        private ITopicClient CreateTopicClient(int index)
        {
            var topicClient = _topicClientFactory.CreateTopicClient(TopicAdapterDescription.ConnectionStrings[index], TopicAdapterDescription.TopicDescription.Path);
            return topicClient;
        }

        public bool TrySend<T>(T message) where T : IBrokeredMessageContainer
        {
            return TrySend(message.BrokeredMessage);
        }

        public bool TrySend(BrokeredMessage message)
        {
            Guard.ArgumentNotNull(message, "message");

            Initialize();

            // increment the starting index to achieve a round-robin effect when talking to multiple topic clients
            // get a local copy of the starting index so that I can iterate it independently of other calls to TrySend
            // and ensure that the round robin effect is touching all of the available clients per call if needed
            var index = IncrementStartingIndex();

            for (int i = 0; i < _topicClients.Length; i++)
            {
                try
                {
                    TopicAdapterDescription.RetryPolicy.Execute(() => _topicClients[index].Send(message));
                    return true;
                }
                catch (Exception ex)
                {
                    if (ex is OperationCanceledException)
                    {
                        // Client entity has been closed or aborted so I am going to replace it
                        _topicClients[index] = CreateTopicClient(index);
                    }

                    IncrementIndex(ref index);

                    OnSendFailed(ex);
                }
            }

            return false;
        }

        private int IncrementStartingIndex()
        {
            lock (_startingIndexLock)
            {
                IncrementIndex(ref _startingIndex);
                return _startingIndex;
            }
        }

        private void IncrementIndex(ref int index)
        {
            index = (++index % TopicAdapterDescription.ConnectionStrings.Length);
        }

        private void OnSendFailed(Exception exception)
        {
            if (SendFailed != null)
            {
                SendFailed(this, new SendFailedEventArgs(exception));
            }
        }

        public event EventHandler<SendFailedEventArgs> SendFailed;

        public TopicAdapterDescription TopicAdapterDescription { get; private set; }
    }
}
