﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.WindowsAzure.StorageClient;
using Nardax.Serialization;

namespace Nardax.Azure.StorageClient
{
    public class BinaryCloudQueueWrapper<T> : CloudQueueWrapperBase, ICloudQueue<T>
    {
        private readonly ISerializer<T, byte[]> _serializer;

        /// <summary>
        /// Initializes a new instance of the CloudQueueWrapper class.
        /// </summary>
        /// <param name="cloudQueue">The cloud queue that this class is wrapping</param>
        /// <param name="serializer">The serializer that is used to convert the message to a byte array</param>
        public BinaryCloudQueueWrapper(CloudQueue cloudQueue, ISerializer<T, byte[]> serializer)
            : base(cloudQueue)
        {
            _serializer = serializer;
        }

        public void AddMessage(T message)
        {
            var serializedValue = _serializer.Serialize(message);
            var cloudQueueMessage = new CloudQueueMessage(serializedValue);
            CloudQueue.AddMessage(cloudQueueMessage);
        }

        public void AddMessage(T message, TimeSpan timeToLive)
        {
            var serializedValue = _serializer.Serialize(message);
            var cloudQueueMessage = new CloudQueueMessage(serializedValue);
            CloudQueue.AddMessage(cloudQueueMessage, timeToLive);
        }

        public ICloudQueueMessage<T> GetMessage()
        {
            var cqm = CloudQueue.GetMessage();
            if (cqm != null)
            {
                var mv = _serializer.Deserialize(cqm.AsBytes);
                return new CloudQueueMessageWrapper<T>(cqm, mv);
            }

            return default(ICloudQueueMessage<T>);
        }

        public ICloudQueueMessage<T> GetMessage(TimeSpan visibilityTimeout)
        {
            var cqm = CloudQueue.GetMessage(visibilityTimeout);
            if (cqm != null)
            {
                var mv = _serializer.Deserialize(cqm.AsBytes);
                return new CloudQueueMessageWrapper<T>(cqm, mv);
            }

            return default(ICloudQueueMessage<T>);
        }

        public IEnumerable<ICloudQueueMessage<T>> GetMessages(int messageCount)
        {
            return (from cqm in CloudQueue.GetMessages(messageCount)
                    select new CloudQueueMessageWrapper<T>(cqm, _serializer.Deserialize(cqm.AsBytes))).ToList();
        }

        public IEnumerable<ICloudQueueMessage<T>> GetMessages(int messageCount, TimeSpan visibilityTimeout)
        {
            return (from cqm in CloudQueue.GetMessages(messageCount, visibilityTimeout)
                    select new CloudQueueMessageWrapper<T>(cqm, _serializer.Deserialize(cqm.AsBytes))).ToList();
        }

        public ICloudQueueMessage<T> PeakMessage()
        {
            var cqm = CloudQueue.PeekMessage();
            if (cqm != null)
            {
                var mv = _serializer.Deserialize(cqm.AsBytes);
                return new CloudQueueMessageWrapper<T>(cqm, mv);
            }

            return default(ICloudQueueMessage<T>);
        }

        public IEnumerable<ICloudQueueMessage<T>> PeakMessages(int messageCount)
        {
            return (from cqm in CloudQueue.PeekMessages(messageCount)
                    select new CloudQueueMessageWrapper<T>(cqm, _serializer.Deserialize(cqm.AsBytes))).ToList();
        }
    }
}