﻿using System;
using System.Configuration;
using Nardax.Data.Configuration;

namespace Nardax.Data
{
    /// <summary>
    /// Provides a factory class for instantiating application-specific retry policies described in the application configuration.
    /// </summary>
    public class RetryPolicyFactory
    {
        public RetryPolicy CreateDefaultRetryPolicy<T>() where T : ITransientErrorDetectionStrategy, new()
        {
            var section = ConfigurationManager.GetSection(NardaxDataSection.SectionName) as NardaxDataSection;

            if (section == null)
            {
                // TODO: add better error message here
                throw new InvalidOperationException();
            }

            return CreateRetryPolicy<T>(section.DefaultRetryPolicyName);
        }
        /// <summary>
        /// Returns an instance of the <see cref="RetryPolicy"/> object initialized for a given policy name.
        /// </summary>
        /// <typeparam name="T">The type implementing the <see cref="ITransientErrorDetectionStrategy"/> interface which is responsible for detecting transient conditions.</typeparam>
        /// <param name="policyName">The name under which a retry policy definition is registered in the application configuration.</param>
        /// <returns>The retry policy initialized from the specified policy definition, or null if the specified policy definition was found.</returns>
        public RetryPolicy CreateRetryPolicy<T>(string policyName) where T : ITransientErrorDetectionStrategy, new()
        {
            Guard.ArgumentNotNullOrEmptyString(policyName, "policyName");

            var section = ConfigurationManager.GetSection(NardaxDataSection.SectionName) as NardaxDataSection;

            if (section == null)
            {
                // TODO: add better error message here
                throw new InvalidOperationException();
            }

            var element = section.RetryPolicies.Get(policyName);

            return CreateRetryPolicy<T>(element);
        }

        /// <summary>
        /// Returns a default policy that implements a fixed retry interval configured with <see cref="RetryPolicy.DefaultClientRetryCount"/> and <see cref="RetryPolicy.DefaultRetryInterval"/> parameters.
        /// </summary>
        public RetryPolicy CreateDefaultFixedRetryPolicy<T>() where T : ITransientErrorDetectionStrategy, new()
        {
            return new RetryPolicy<T>(RetryPolicy.DefaultClientRetryCount, RetryPolicy.DefaultRetryInterval);
        }

        /// <summary>
        /// Returns a default policy that implements a progressive retry interval configured with <see cref="RetryPolicy.DefaultClientRetryCount"/>, <see cref="RetryPolicy.DefaultRetryInterval"/> and <see cref="RetryPolicy.DefaultRetryIncrement"/> parameters.
        /// </summary>
        public RetryPolicy CreateDefaultProgressiveRetryPolicy<T>() where T : ITransientErrorDetectionStrategy, new()
        {
            return new RetryPolicy<T>(RetryPolicy.DefaultClientRetryCount, RetryPolicy.DefaultRetryInterval, RetryPolicy.DefaultRetryIncrement);
        }

        /// <summary>
        /// Returns a default policy that implements a random exponential retry interval configured with <see cref="RetryPolicy.DefaultClientRetryCount"/>, <see cref="RetryPolicy.DefaultMinBackoff"/>, <see cref="RetryPolicy.DefaultMaxBackoff"/> and <see cref="RetryPolicy.DefaultClientBackoff"/> parameters.
        /// </summary>
        public RetryPolicy CreateDefaultExponentialRetryPolicy<T>() where T : ITransientErrorDetectionStrategy, new()
        {
            return new RetryPolicy<T>(RetryPolicy.DefaultClientRetryCount, RetryPolicy.DefaultMinBackoff, RetryPolicy.DefaultMaxBackoff, RetryPolicy.DefaultClientBackoff);
        }

        /// <summary>
        /// Returns an instance of the <see cref="RetryPolicy"/> object based on the parameters in the current policy definition object.
        /// </summary>
        /// <typeparam name="T">The type implementing the <see cref="ITransientErrorDetectionStrategy"/> interface which is responsible for detecting transient conditions.</typeparam>
        /// <returns>The retry policy initialized from the current policy definition, or a null reference if method was unable to detect the type of retry method.</returns>
        private static RetryPolicy<T> CreateRetryPolicy<T>(RetryPolicyElement element) where T : ITransientErrorDetectionStrategy, new()
        {
            RetryPolicy<T> retryPolicy;

            // Now it's time to make a decision as to what retry policy is required. First, check for any back-off parameters.
            if (element.MinBackoff != 0 || element.MaxBackoff != 0 || element.DeltaBackoff != 0)
            {
                retryPolicy = new RetryPolicy<T>(element.MaxRetryCount, TimeSpan.FromMilliseconds(element.MinBackoff), TimeSpan.FromMilliseconds(element.MaxBackoff), TimeSpan.FromMilliseconds(element.DeltaBackoff));
            }
            // Check if RetryIncrement was specified - this indicates that progressive delay retry policy is requested.
            else if (element.RetryIncrement != 0)
            {
                retryPolicy = new RetryPolicy<T>(element.MaxRetryCount, TimeSpan.FromMilliseconds(element.RetryInterval), TimeSpan.FromMilliseconds(element.RetryIncrement));
            }
            // Assume it's a fixed interval retry policy.
            else
            {
                retryPolicy = new RetryPolicy<T>(element.MaxRetryCount, TimeSpan.FromMilliseconds(element.RetryInterval));
            }

            return retryPolicy;
        }
    }
}
