﻿using System;

namespace Nardax
{
    public static class DateTimeExtensions
    {
        public static DateTime Floor(this DateTime dateTime, TimeSpan timeSpan)
        {
            var ticks = dateTime.Ticks / timeSpan.Ticks;
            return new DateTime(ticks * timeSpan.Ticks);
        }

        public static DateTime Round(this DateTime dateTime, TimeSpan timeSpan)
        {
            var ticks = (dateTime.Ticks + (timeSpan.Ticks / 2) + 1) / timeSpan.Ticks;
            return new DateTime(ticks * timeSpan.Ticks);
        }

        public static DateTime Ceiling(this DateTime dateTime, TimeSpan timeSpan)
        {
            var ticks = (dateTime.Ticks + timeSpan.Ticks - 1) / timeSpan.Ticks;
            return new DateTime(ticks * timeSpan.Ticks); 
        }

        public static DateTime Min(this DateTime dateTime, DateTime value)
        {
            return dateTime <= value ? dateTime : value;
        }

        public static DateTime Max(this DateTime dateTime, DateTime value)
        {
            return dateTime >= value ? dateTime : value;
        }

        public static int SecondsFrom(this DateTime dateTime, DateTime from)
        {
            var timeSpan = (dateTime - from);
            return Convert.ToInt32(timeSpan.TotalSeconds);
        }

        /// <summary>
        /// Returns the number of seconds from Jan 1st, 1970 to the source date time.
        /// </summary>
        public static int SecondsFromUnixEpoch(this DateTime dateTime)
        {
            return dateTime.SecondsFrom(new DateTime(1970, 1, 1));
        }

        /// <summary>
        /// Returns the total number of milliseconds to get from midnight of the current date to the current time.
        /// </summary>
        public static double MillisecondOfDay(this DateTime dateTime)
        {
            return dateTime.TimeOfDay.TotalMilliseconds;
        }

        /// <summary>
        /// Resets the time portion to Midnight while preserving the date.
        /// </summary>
        public static DateTime Midnight(this DateTime dateTime)
        {
            return dateTime.Date;
        }

        /// <summary>
        /// Resets the time portion to Noon while preserving the date.
        /// </summary>
        public static DateTime Noon(this DateTime dateTime)
        {
            return new DateTime(dateTime.Year, dateTime.Month, dateTime.Day, 12, 0, 0);
        }

        /// <summary>
        /// Resets the time portion to the provided hour and minute. Second and millisecond are set to zero.
        /// </summary>
        public static DateTime SetTime(this DateTime dateTime, int hour, int minute)
        {
            return SetTime(dateTime, hour, minute, 0, 0);
        }

        /// <summary>
        /// Resets the time portion to the provided hour, minute and second. Millisecond is set to zero.
        /// </summary>
        public static DateTime SetTime(this DateTime dateTime, int hour, int minute, int second)
        {
            return SetTime(dateTime, hour, minute, second, 0);
        }

        /// <summary>
        /// Resets the time portion to the provided hour, minute, second and millisecond.
        /// </summary>
        public static DateTime SetTime(this DateTime dateTime, int hour, int minute, int second, int millisecond)
        {
            return new DateTime(dateTime.Year, dateTime.Month, dateTime.Day, hour, minute, second, millisecond);
        }
    }
}
