﻿using System;
using System.Diagnostics;
using System.Threading;

namespace Nardax
{
    public class RetryPolicy : IRetryPolicy
    {
        public RetryPolicy(ITransientErrorDetectionStrategy transientErrorDetectionStrategy, IRetryStrategy retryStrategy)
        {
            Guard.ArgumentNotNull(transientErrorDetectionStrategy, "transientErrorDetectionStrategy");
            Guard.ArgumentNotNull(retryStrategy, "retryStrategy");

            TransientErrorDetectionStrategy = transientErrorDetectionStrategy;
            RetryStrategy = retryStrategy;
        }

        public void Execute(Action method)
        {
            Execute(method, null, null);
        }

        public void Execute(Action method, Stopwatch stopwatch)
        {
            Execute(method, stopwatch, null);
        }

        public void Execute(Action method, Action<int, Exception, TimeSpan> onRetry)
        {
            Execute(method, null, onRetry);
        }

        public void Execute(Action method, Stopwatch stopwatch, Action<int, Exception, TimeSpan> onRetry)
        {
            Execute(() =>
            {
                method();

                // return the default object here so that the method signature will match the Execute(Func<TResult> method).
                return default(object);
            }, stopwatch, onRetry);
        }

        public TResult Execute<TResult>(Func<TResult> method)
        {
            return Execute(method, null, null);
        }

        public TResult Execute<TResult>(Func<TResult> method, Stopwatch stopwatch)
        {
            return Execute(method, stopwatch, null);
        }

        public TResult Execute<TResult>(Func<TResult> method, Action<int, Exception, TimeSpan> onRetry)
        {
            return Execute(method, null, onRetry);
        }

        public TResult Execute<TResult>(Func<TResult> method, Stopwatch stopwatch, Action<int, Exception, TimeSpan> onRetry)
        {
            Guard.ArgumentNotNull(method, "method");

            var retryCount = 0;
            TimeSpan delay;
            Exception lastError;
            ShouldRetry shouldRetry = RetryStrategy.GetShouldRetry();

            while (true)
            {
                if (stopwatch != null)
                {
                    stopwatch.Restart();
                }

                try
                {
                    return method();
                }
                catch (Exception ex)
                {
                    lastError = ex;

                    if (!(TransientErrorDetectionStrategy.IsTransient(lastError) && shouldRetry(retryCount++, lastError, out delay)))
                    {
                        throw;
                    }
                }
                finally
                {
                    if (stopwatch != null)
                    {
                        stopwatch.Stop();
                    }
                }

                // Perform an extra check in the delay interval. Should prevent from accidentally ending up with the value of -1 that will 
                // block a thread indefinitely. In addition, any other negative numbers will cause an ArgumentOutOfRangeException fault 
                // that will be thrown by Thread.Sleep.
                if (delay.TotalMilliseconds < 0)
                {
                    delay = TimeSpan.Zero;
                }

                if (onRetry != null)
                {
                    onRetry(retryCount, lastError, delay);
                }

                OnRetrying(retryCount, lastError, delay);

                if (retryCount > 1 || !RetryStrategy.FastFirstRetry)
                {
                    Thread.Sleep(delay);
                }
            }
        }

        /// <summary>
        /// Notifies the subscribers whenever a retry condition is encountered.
        /// </summary>
        /// <param name="retryCount">The current retry attempt count.</param>
        /// <param name="lastError">The exception which caused the retry conditions to occur.</param>
        /// <param name="delay">The delay indicating how long the current thread will be suspended for before the next iteration will be invoked.</param>
        protected virtual void OnRetrying(int retryCount, Exception lastError, TimeSpan delay)
        {
            if (Retrying != null)
            {
                Retrying(this, new RetryingEventArgs(retryCount, delay, lastError));
            }
        }

        public event EventHandler<RetryingEventArgs> Retrying;

        public ITransientErrorDetectionStrategy TransientErrorDetectionStrategy { get; private set; }

        public IRetryStrategy RetryStrategy { get; private set; }
    }
}